#!/usr/bin/python3
#
# autoblockchainify — Turn a directory into a GIT Blockchain
#
# Copyright (C) 2019-2021 Marcel Waldvogel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# Sending and receiving mail

import signale
import os
import re
import subprocess
import threading
import time
from datetime import datetime, timedelta
from imaplib import IMAP4
from pathlib import Path
from smtplib import SMTP
from time import gmtime, strftime

import pygit2 as git

import autoblockchainify.config

logging = signale.Signale({"scope": "mail"})
serialize_receive = threading.Lock()
serialize_create = threading.Lock()


def split_host_port(host, default_port):
    if ':' in host:
        host, port = host.split(':', 1)
        return (host, int(port))
    else:
        return (host, default_port)


def send(body, subject='Stamping request', to=None):
    # Does not work in unittests if assigned in function header
    # (are bound too early? At load time instead of at call time?)
    if to is None:
        to = autoblockchainify.config.arg.stamper_to
    (host, port) = split_host_port(
        autoblockchainify.config.arg.stamper_smtp_server, 587)
    with SMTP(host, port=port) as smtp:
        smtp.starttls()
        smtp.login(autoblockchainify.config.arg.stamper_username,
                   autoblockchainify.config.arg.stamper_password)
        frm = autoblockchainify.config.arg.stamper_own_address
        date = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
        msg = """From: %s
To: %s
Date: %s
Subject: %s

%s""" % (frm, to, date, subject, body)
        smtp.sendmail(frm, to, msg)
        logging.complete("Timestamping request mailed")


def extract_pgp_body(body):
    try:
        body = str(body, 'ASCII')
    except TypeError as t:
        logging.warning("Conversion error for message body: %s" % t)
        return None
    lines = body.splitlines()
    start = None
    for i in range(0, len(lines)):
        if lines[i] == '-----BEGIN PGP SIGNED MESSAGE-----':
            logging.debug("Found start at line %d: %s" % (i, lines[i]))
            start = i
            break
    else:
        return None

    end = None
    for i in range(start, len(lines)):
        if lines[i] == '-----END PGP SIGNATURE-----':
            logging.debug("Found end at line %d: %s" % (i, lines[i]))
            end = i
            break
    else:
        return None
    return lines[start:end + 1]


def save_signature(bodylines, logfile):
    logging.xdebug("save_signature()")
    repo = autoblockchainify.config.arg.repository
    ascfile = Path(repo, 'pgp-timestamp.sig')
    with ascfile.open(mode='w') as f:
        f.write('\n'.join(bodylines) + '\n')
        # Change will be picked up by next check for directory modification
    logfile.unlink()  # Mark as reply received, no need for resumption


def maybe_decode(s):
    """Decode, if it is not `None`"""
    if s is None:
        return s
    else:
        return s.decode('ASCII')


def body_signature_correct(bodylines, stat):
    body = '\n'.join(bodylines)
    logging.debug("Bodylines", suffix=body)
    # Cannot use Python gnupg wrapper: Requires GPG 1.x to verify
    # Copy env for gnupg without locale
    env = {}
    for k in os.environ:
        if not k.startswith('LC_'):
            env[k] = os.environ[k]
    env['LANG'] = 'C'
    env['TZ'] = 'UTC'
    res = subprocess.run(['gpg1', '--pgp2', '--verify'],
                         env=env, input=body.encode('ASCII'),
                         stderr=subprocess.PIPE)
    stderr = maybe_decode(res.stderr)
    logging.complete(stderr)
    if res.returncode != 0:
        logging.error("gpg1 return code %d (%r)" % (res.returncode, stderr))
        return False
    if '\ngpg: Good signature' not in stderr:
        logging.error("Missing good signature (%r)" % stderr)
        return False
    if not stderr.startswith('gpg: Signature made '):
        logging.error("No signature made (%r)" % stderr)
        return False
    if not ((' key ID %s\n' % autoblockchainify.config.arg.stamper_keyid)
            in stderr):
        logging.error("Signature by wrong KeyID (%r)" % stderr)
        return False
    try:
        sigtime = datetime.strptime(stderr[24:48], "%b %d %H:%M:%S %Y %Z")
        logging.xdebug("Extracted date %r" % sigtime)
    except ValueError:
        logging.error("Illegal signature date format %r (%r)" %
                      (stderr[24:48], stderr))
        return False
    if sigtime > datetime.utcnow() + timedelta(seconds=30):
        logging.error("Signature time %s lies more than 30 seconds in the future"
                      % sigtime)
        return False
    modtime = datetime.utcfromtimestamp(stat.st_mtime)
    if sigtime < modtime - timedelta(seconds=30):
        logging.error("Signature time %s is more than 30 seconds before "
                      "file modification time %s"
                      % (sigtime, modtime))
        return False
    return True


def verify_body_and_save_signature(body, stat, logfile, msgno):
    bodylines = extract_pgp_body(body)
    if bodylines is None:
        logging.error("No body lines")
        return False

    res = body_contains_file(bodylines, logfile)
    if res is None:
        logging.error("File contents not in message %s" % msgno)
        return False
    else:
        (before, after) = res
        logging.debug("Message wrapped in %d lines before, %d after" %
                      (before, after))
        if before > 20 or after > 20:
            logging.error("Too many lines added by the PGP Timestamping Server"
                          " before (%d)/after (%d) our contents" % (before, after))
            return False

    if not body_signature_correct(bodylines, stat):
        logging.error("Body signature incorrect")
        return False

    save_signature(bodylines, logfile)
    return True


def body_contains_file(bodylines, logfile):
    if bodylines is None:
        return None
    linesbefore = 0
    with logfile.open(mode='r') as f:
        # A few empty/comment lines at the beginning
        firstline = f.readline().rstrip()
        for i in range(len(bodylines)):
            if bodylines[i] == firstline:
                break
            elif bodylines[i] == '' or bodylines[i][0] in '#-':
                linesbefore += 1
            else:
                return None
        # Now should be contiguous
        i += 1
        for line in f:
            if bodylines[i] != line.rstrip():
                return None
            i += 1
        # Now there should only be empty lines and a PGP signature
        linesafter = len(bodylines) - i
        i += 1
        while bodylines[i] == '':
            i += 1
        if bodylines[i] != '-----BEGIN PGP SIGNATURE-----':
            return None
        # No further line starting with '-'
        for i in range(i + 1, len(bodylines) - 1):
            if bodylines[i] != '' and bodylines[i][0] == '-':
                return None
        return (linesbefore, linesafter)


def file_unchanged(stat, logfile):
    try:
        cur_stat = logfile.stat()
        unchanged = (cur_stat.st_mtime == stat.st_mtime
                     and cur_stat.st_ino == stat.st_ino)
        logging.xdebug("%r unchanged: %r" % (logfile, unchanged))
        return unchanged
    except FileNotFoundError:
        logging.xdebug("%r unchanged: missing" % logfile)
        return False


def imap_idle(imap, stat, logfile):
    while True:
        imap.send(b'%s IDLE\r\n' % (imap._new_tag()))
        logging.pause("IMAP idling")
        line = imap.readline().strip()
        logging.debug("IMAP IDLE → %s" % line)
        if line != b'+ idling':
            logging.error("IMAP IDLE unsuccessful")
            return False
        # Wait for new message
        while file_unchanged(stat, logfile):
            line = imap.readline().strip()
            if line == b'' or line.startswith(b'* BYE '):
                return False
            match = re.match(r'^\* ([0-9]+) EXISTS$', str(line, 'ASCII'))
            if match:
                logging.success("You have new mail %s!"
                                % match.group(1).encode('ASCII'))
                # Stop idling
                imap.send(b'DONE\r\n')
                if check_for_stamper_mail(imap, stat, logfile) is True:
                    logging.xdebug("Returning from success after idling")
                    return False
                break  # Restart IDLE command
            # Otherwise: Uninteresting untagged response, continue idling
        logging.stop("Next mail sent, giving up waiting on now-old reply")


def check_for_stamper_mail(imap, stat, logfile):
    # See `--no-dovecot-bug-workaround`:
    query = ('FROM', '"%s"' % autoblockchainify.config.arg.stamper_from,
             'UNSEEN',
             'LARGER', str(stat.st_size),
             'SMALLER', str(stat.st_size + 16384))
    logging.debug("IMAP SEARCH " + (' '.join(query)))
    (typ, msgs) = imap.search(None, *query)
    logging.info("IMAP SEARCH → %s, %s" % (typ, msgs))
    if len(msgs) == 1 and len(msgs[0]) > 0:
        mseq = msgs[0].replace(b' ', b',')
        (typ, contents) = imap.fetch(mseq, 'BODY[TEXT]')
        logging.debug("IMAP FETCH → %s (%d)" % (typ, len(contents)))
        remaining_msgids = mseq.split(b',')
        for m in contents:
            if m != b')':
                msgid = remaining_msgids[0]
                remaining_msgids = remaining_msgids[1:]
                logging.debug("IMAP FETCH BODY (%s) → %s…" %
                              (msgid, m[1][:20]))
                if verify_body_and_save_signature(m[1], stat, logfile, msgid):
                    logging.success(
                        "Successful answer in message %s; deleting" % msgid)
                    imap.store(msgid, '+FLAGS', '\\Deleted')
                    return True
    return False


def wait_for_receive(logfile):
    try:
        logging.start("wait_for_receive", level=signale.XDEBUG,
                      suffix="Threads: " + str(threading.enumerate()))
        with serialize_receive:
            if not logfile.is_file():
                logging.warning("Logfile vanished, should not happen")
                return
            stat = logfile.stat()
            logging.debug("Timestamp revision file is from %d" % stat.st_mtime)
            (host, port) = split_host_port(
                autoblockchainify.config.arg.stamper_imap_server, 143)
            with IMAP4(host=host, port=port) as imap:
                imap.starttls()
                imap.login(autoblockchainify.config.arg.stamper_username,
                           autoblockchainify.config.arg.stamper_password)
                imap.select('INBOX')
                if not check_for_stamper_mail(imap, stat, logfile):
                    # No existing message found, wait for more incoming messages
                    # and process them until definitely okay or giving up for good
                    if 'IDLE' in imap.capabilities:
                        imap_idle(imap, stat, logfile)
                    else:
                        logging.warning(
                            "IMAP server does not support IDLE; polling instead")
                        # Poll every minute, for 10 minutes;
                        # see description for `async_email_timestamp` below.
                        for _ in range(10):
                            time.sleep(60)
                            if check_for_stamper_mail(imap, stat, logfile):
                                logging.success("Polling found matching mail")
                                return
                        logging.stop("No response received, giving up")
        logging.success("Returning from mail thread", level=signale.XDEBUG)
    except Exception:
        logging.exception("Unhandled exception in mail thread")


def modified_in(file, wait):
    """Has `logfile` not been modified in ~`wait` seconds?
    Non-existent file is considered to *not* fulfill this."""
    try:
        stat = file.stat()
        mtime = datetime.utcfromtimestamp(stat.st_mtime)
        now = datetime.utcnow()
        logging.xdebug("modified_in(%s, %s): mtime %s, now %s" %
                       (file, wait, mtime, now))
        return mtime + wait >= now
    except FileNotFoundError:
        logging.xdebug("modified_in: %s not found" % file)
        return False


# * `resume=True`: Run once at startup, to wait for a possibly pending
#   outstanding mail reply, indicated by the presence of the logfile.
#   Cases:
#   1. Pending email (indicated by logfile)
#      - Wait for reply (should already have arrived or in the next 5 minutes)
#      - Abort when new logfile arrives
#   2. No pending mail
#      - Do nothing, wait for next commit
# * `resume=False`: Run at least once every
#   `commit_interval * force_after_intervals` (± some jitter).
#   sigfile may be generated up to about 5 minutes later (mail
#   timestamper processes mails on every full 5-minute interval).
#   Cases:
#   1. First commit (no sigfile yet)
#      - Send mail unconditionally
#      - Wait for reply, abort when new logfile arrives
#   2. Only one forced commit every force interval:
#      - Send mail unconditionally on every commit
#      - Wait for reply, abort when new logfile arrives
#   3. Frequent changes (several of the commit_intervals):
#      - Previous sigfile should already have been committed
#        (as we are called only when a commit has just been made)
#      - Send mail when mail reply would be received no
#        earlier than the previous sigfile date plus force_interval
#      - Wait for reply, abort when new logfile arrives
#   4. Previous mail has not been replied to (mail problem or too frequent)
#      - Force send after force_interval anyway
#      - Wait for reply, abort when new logfile arrives
#   Solution:
#   - Trigger new mail 3…4 minutes before sigfile time+force_interval
#     (assuming that force_interval is a multiple of 5 minutes; if
#     not, mails might be up to 3…4 minutes more frequent than
#     force_interval, possibly causing more than one mail per
#     forced commit, i.e., some work by the PGP timestamper is wasted)
#   - The minimum force_interval has to greater than 4+5 minutes
#   - Do not trigger new mail if logfile exists and has been written to in the
#     past 4+5-epsilon minutes
# Note: `wait` is almost 1 commit_interval shorter than forced_interval
# (see `do_commit()`.)
def needs_timestamp(log=False):
    if not autoblockchainify.config.arg.stamper_own_address:
        if log:
            logging.debug("Timestamping by mail not configured")
        return False
    path = autoblockchainify.config.arg.repository
    logfile = Path(path, 'pgp-timestamp.tmp')
    sigfile = Path(path, 'pgp-timestamp.sig')
    sigfile_interval = (autoblockchainify.config.arg.commit_interval
                        * autoblockchainify.config.arg.force_after_intervals
                        - timedelta(minutes=4))
    if logfile.is_file() and modified_in(logfile, timedelta(minutes=4+5)):
        if log:
            logging.stop("Logfile more recent than 4+5 minutes, skipping")
        return False
    if not sigfile.is_file() or not modified_in(sigfile, sigfile_interval):
        return True
    else:
        if log:
            logging.debug("Sigfile too fresh")
        return False


def async_email_timestamp(resume=False):
    """If called with `resume=True`, tries to resume waiting for the mail"""
    logging.xdebug("async_email_timestamp(%r)" % resume)
    path = autoblockchainify.config.arg.repository
    repo = git.Repository(path)
    if repo.head_is_unborn:
        logging.stop(
            "Cannot timestamp by email yet: repository without commits")
        return
    head = repo.head
    logfile = Path(path, 'pgp-timestamp.tmp')
    if resume:
        if not logfile.is_file():
            logging.stop("Not resuming mail timestamp: No pending mail reply")
            return
        with logfile.open() as f:
            contents = f.read()
        logging.xdebug("Resuming with logfile contents: %r" % contents)
        if len(contents) < 40:
            logging.stop("Not resuming mail timestamp: No revision info")
            return
        threading.Thread(target=wait_for_receive, args=(logfile,), name="mail_resume",
                         daemon=True).start()
    else:  # Fresh request
        # No recent attempts or results for mail timestamping
        if needs_timestamp(log=True):
            new_rev = ("git commit %s\nTimestamp requested at %s\n" %
                       (head.target.hex,
                        strftime("%Y-%m-%d %H:%M:%S UTC", gmtime())))
            logging.xdebug("Creating logfile with: %r" % new_rev)
            with serialize_create:
                with logfile.open('w') as f:
                    f.write(new_rev)
                send(new_rev)
            threading.Thread(target=wait_for_receive, args=(logfile,), name="mail",
                             daemon=True).start()



#!/usr/bin/python3
#
# autoblockchainify — Turn a directory into a GIT Blockchain
#
# Copyright (C) 2019-2021 Marcel Waldvogel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# Set up the daemon


import signale
import subprocess
from pathlib import Path

import autoblockchainify.commit
import autoblockchainify.config
import autoblockchainify.version


logging = signale.Signale({"scope": "daemon"})
try:
    # Aligned output, if supported by this version of Signale
    signale.set_align(8, 8)
except AttributeError:
    pass


def finish_setup(arg):
    # Create git repository, if necessary and set user name/email
    repo = autoblockchainify.config.arg.repository
    Path(repo).mkdir(parents=True, exist_ok=True)
    if not Path(repo, '.git').is_dir():
        subprocess.run(['git', 'init'], cwd=repo, check=True)
        # Default identity is forced onto new repositories only;
        # otherwise, the default is not to change the user settings.
        if arg.identity is None:
            arg.identity = 'Autoblockchainify <autoblockchainify@localhost>'
        # Ignore 'pgp-timestamper.rev'
        if not Path(repo, '.gitignore').is_file():
            with open(Path(repo, '.gitignore'), 'a') as ignore:
                ignore.write('''pgp-timestamp.tmp
.gnupg/
.ssh/
''')
    if arg.identity is not None:
        logging.start("Initializing new repo with user info")
        (name, mail) = arg.identity.split(' <')
        subprocess.run(['git', 'config', 'user.name', name],
                       cwd=repo, check=True)
        subprocess.run(['git', 'config', 'user.email', mail[:-1]],
                       cwd=repo, check=True)


def run():
    autoblockchainify.config.get_args()
    finish_setup(autoblockchainify.config.arg)
    # Try to resume waiting for a PGP Timestamping Server reply, if any
    if autoblockchainify.config.arg.stamper_own_address:
        logging.pending("possibly resuming cross-timestamping by mail")
        autoblockchainify.mail.async_email_timestamp(resume=True)
    autoblockchainify.commit.loop()



#!/usr/bin/python3
#
# autoblockchainify — Turn any directory into a GIT Blockchain
#
# Copyright (C) 2019-2021 Marcel Waldvogel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# Configuration handling

import argparse
import configargparse
import datetime
import signale
import os
import sys
import random
import deltat

import autoblockchainify.version


logging = signale.Signale({"scope": "config"})


def get_args(args=None, config_file_contents=None):
    global arg
    # Config file in /etc or the program directory
    parser = configargparse.ArgumentParser(
        auto_env_var_prefix="autoblockchainify_",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="autoblockchainify — Turn any directory into a GIT Blockchain.",
        default_config_files=['/etc/autoblockchainify.conf',
                              os.path.join(os.getenv('HOME'), 'autoblockchainify.conf')])

    # General
    parser.add_argument('--config-file', '-c',
                        is_config_file=True,
                        help="config file path")
    parser.add_argument('--debug-level',
                        default='INFO',
                        help="""amount of debug output: WARN, INFO, or DEBUG.
                            Debug levels for specific loggers can also be
                            specified using 'name=level'. Valid logger names:
                            `config`, `daemon`, `commit` (incl. requesting
                            timestamps), `gnupg`, `mail` (interfacing with PGP
                            Timestamping Server).  Example: `DEBUG,gnupg=INFO`
                            sets the default debug level to DEBUG, except for
                            `gnupg`.""")
    parser.add_argument('--version',
                        action='version', version=autoblockchainify.version.VERSION)

    # Identity
    # Default is applied in daemon.py:finish_setup()
    parser.add_argument('--identity',
                        help="""'Full Name <email@address>' for tagging
                        GIT commits.
                        Default (will only be applied to new repositories):
                        'Autoblockchainify <autoblockchainify@localhost>'.
                        An explicit value (non-default) will always update
                        the current GIT config.""")

    # Stamping
    parser.add_argument('--commit-interval',
                        default='10m',
                        help="how often to commit, if there have been changes")
    parser.add_argument('--commit-offset',
                        help="""when to commit within that interval; e.g. after
                            7m19.3s. Default: Random choice in the interval,
                            avoiding the first/last 5%.""")
    parser.add_argument('--force-after-intervals',
                        type=int,
                        default=6,
                        help="""After how many intervals to force a commit (and
                            request a timestamp by email, if configured).""")
    parser.add_argument('--repository',
                        default='.',
                        help="""path to the GIT repository (default '.')""")
    parser.add_argument('--zeitgitter-servers',
                        default='diversity gitta',
                        help="""any number of space-separated
                             Zeitgitter servers of the form
                             `[<branch>=]<server>`. The server name will
                             be passed with `--server` to `git timestamp`,
                             the (optional) branch name with `--branch`.""")
    parser.add_argument('--zeitgitter-sleep',
                        default='0s',
                        help="""Delay between cross-timestamping for the
                             different timestampers""")

    # Pushing
    parser.add_argument('--push-repository',
                        default='',
                        help="""Space-separated list of repositores to push to;
                            setting this enables automatic push""")
    parser.add_argument('--push-branch',
                        default='*',
                        help="""Space-separated list of branches to push.
                            `*` means all, as `--all` is eaten by ConfigArgParse""")

    # PGP Digital Timestamper interface
    parser.add_argument('--stamper-own-address', '--mail-address', '--email-address',
                        help="""our email address; enables
                            cross-timestamping from the PGP timestamper""")
    parser.add_argument('--stamper-keyid', '--external-pgp-timestamper-keyid',
                        default="70B61F81",
                        help="PGP key ID to obtain email cross-timestamps from")
    parser.add_argument('--stamper-to', '--external-pgp-timestamper-to',
                        default="clear@stamper.itconsult.co.uk",
                        help="""destination email address
                            to obtain email cross-timestamps from""")
    parser.add_argument('--stamper-from', '--external-pgp-timestamper-from',
                        default="mailer@stamper.itconsult.co.uk",
                        help="""email address used by PGP timestamper
                            in its replies""")
    parser.add_argument('--stamper-smtp-server', '--smtp-server',
                        help="""SMTP server to use for
                            sending mail to PGP Timestamper""")
    parser.add_argument('--stamper-imap-server', '--imap-server',
                        help="""IMAP server to use for
                            receiving mail from PGP Timestamper""")
    parser.add_argument('--stamper-username', '--mail-username',
                        help="""username to use for IMAP and SMTP
                            (default from `--stamper-own-address`)""")
    parser.add_argument('--stamper-password', '--mail-password',
                        help="password to use for IMAP and SMTP")
    parser.add_argument('--no-dovecot-bug-workaround', action='store_true',
                        help="""Some Dovecot mail server seem unable to match
                            the last char of an email address in an IMAP
                            SEARCH, so this cuts off the last char from
                            `stamper-from`. Should not impact other mail
                            servers.""")

    arg = parser.parse_args(
        args=args, config_file_contents=config_file_contents)

    for level in str(arg.debug_level).split(','):
        if '=' in level:
            (logger, lvl) = level.split('=', 1)
        else:
            logger = None  # Root logger
            lvl = level
        try:
            lvl = int(lvl)
            lvl = signale.WARNING - lvl * (signale.WARNING - signale.INFO)
        except ValueError:
            pass
        signale.set_threshold(logger, lvl)

    if arg.stamper_username is None:
        arg.stamper_username = arg.stamper_own_address

    if arg.force_after_intervals < 2:
        sys.exit("--force-after-intervals must be >= 2")

    arg.commit_interval = deltat.parse_time(arg.commit_interval)
    if arg.stamper_own_address is None:
        if arg.commit_interval < datetime.timedelta(minutes=1):
            sys.exit("--commit-interval may not be shorter than 1m")
    else:
        if arg.commit_interval*arg.force_after_intervals < datetime.timedelta(minutes=10):
            sys.exit("--commit-interval times --force-after-intervals may "
                     "not be shorter than 10m when using the (mail-based) "
                     "PGP Digital Timestamping Service")

    if arg.commit_offset is None:
        # Avoid the seconds around the full interval, to avoid clustering
        # with other system activity.
        arg.commit_offset = arg.commit_interval * random.uniform(0.05, 0.95)
        logging.info("Chose --commit-offset %s" % arg.commit_offset)
    else:
        arg.commit_offset = deltat.parse_time(arg.commit_offset)
    if arg.commit_offset < datetime.timedelta(seconds=0):
        sys.exit("--commit-offset must be positive")
    if arg.commit_offset >= arg.commit_interval:
        sys.exit("--commit-offset must be less than --commit-interval")

    arg.zeitgitter_sleep = deltat.parse_time(arg.zeitgitter_sleep)

    # Work around ConfigArgParse list bugs by implementing lists ourselves
    arg.zeitgitter_servers = arg.zeitgitter_servers.split()
    arg.push_repository = arg.push_repository.split()

    # and working around the problem that values cannot start with `-`.
    if arg.push_branch == '*':
        arg.push_branch = ['--all']
    else:
        arg.push_branch = arg.push_branch.split()

    if not arg.no_dovecot_bug_workaround:
        arg.stamper_from = arg.stamper_from[:-1]  # See help text

    logging.success("Settings applied: %s" % str(arg), level=signale.DEBUG)
    return arg



#!/usr/bin/python3
import os
import re
import subprocess

import setuptools


def extract_version(filename):
    with open(filename, 'r') as fh:
        for line in fh:
            match = re.match('''VERSION\s*=\s*["']([-_.0-9a-z]+)(\+?)["']''', line)
            if match:
                if match[2] == '':
                    return match[1]
                else:
                    return match[1] + '.post'
    exit("Cannot extract version number from %s" % filename)


def describe_or_extract_version(filename):
    if 'FORCE_VERSION' in os.environ:
        return os.environ['FORCE_VERSION']
    ret = subprocess.run(['git', 'describe'], capture_output=True, text=True)
    if ret.returncode != 0:
        return extract_version(filename)
    else:
        match = re.match('^v?([0-9]+.[0-9]+.[0-9]+)(-([0-9]+))?', ret.stdout)
        if match:
            if match[3] is None:
                return match[1]
            else:
                return match[1] + '.post' + match[3]
        else:
            return extract_version(filename)


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="autoblockchainify",
    version=describe_or_extract_version('autoblockchainify/version.py'),
    author="Marcel Waldvogel",
    author_email="marcel.waldvogel@trifence.ch",
    description="Turn a directory into a git-based Blockchain",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/zeitgitter/autoblockchainify",
    license='AGPLv3',
    packages=setuptools.find_packages(),
    install_requires=['pygit2', 'configargparse~=1.2', 'deltat>=1.0.1',
        'setuptools', 'git-timestamp>=1.0.4', 'signale-logging>=0.5.3'],
    package_data={'autoblockchainify': ['sample.conf', 'web/*']},
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'autoblockchainify=autoblockchainify.daemon:run',
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: No Input/Output (Daemon)",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Software Development :: Version Control :: Git",
        "Topic :: Security",
    ],
)



#!/usr/bin/python3
#
# autoblockchainify — Turn a directory into a GIT Blockchain
#
# Copyright (C) 2019-2021 Marcel Waldvogel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# Committing to git and obtaining timestamps

from datetime import datetime, timezone
import signale
from pathlib import Path
import subprocess
import sys
import threading
import time
import traceback

import pygit2 as git

import autoblockchainify.config
import autoblockchainify.mail

logging = signale.Signale({"scope": "commit"})


def push_upstream(repo, to, branches):
    logging.pending("Pushing to %s" % (['git', 'push', to] + branches))
    ret = subprocess.run(['git', 'push', to] + branches,
                         cwd=repo)
    if ret.returncode != 0:
        logging.error("'git push %s %s' failed" % (to, ' '.join(branches)))


def cross_timestamp(repo, options, server):
    ret = subprocess.run(['git', 'timestamp'] + options, cwd=repo)
    if ret.returncode != 0:
        logging.error("git timestamp %s failed" % (' '.join(options)))
    else:
        logging.success("Timestamped against %s" % server)


def has_user_changes(repo):
    """Check whether there are uncommitted changes, i.e., whether
    `git status -z` has any output. A modification of only `pgp-timestamp.sig`
    is ignored, as it is neither necessary nor desirable to trigger on it:
    (a) our own timestamp is not really needed on it and
    (b) it would cause an unnecessary second timestamp per idle force period."""
    ret = subprocess.run(['git', 'status', '-z'],
                         cwd=repo, capture_output=True, check=True)
    return len(ret.stdout) > 0 and ret.stdout != b' M pgp-timestamp.sig\0'


def pending_merge(repo):
    """Check whether there is a pending merge."""
    return Path(repo, '.git', 'MERGE_HEAD').is_file()


def commit_current_state(repo):
    """Force a commit; will be called only if a commit has to be made.
    I.e., if there really are changes or the force duration has expired."""
    now = datetime.now(timezone.utc)
    nowstr = now.strftime('%Y-%m-%d %H:%M:%S UTC')
    subprocess.run(['git', 'add', '.'],
                   cwd=repo, check=True)
    subprocess.run(['git', 'commit', '--allow-empty',
                    '-m', "🔗 Autoblockchainify data as of " + nowstr],
                   cwd=repo, check=True)


def head_older_than(repo, duration):
    """Check whether the last commit is older than `duration`.
    Unborn HEAD is *NOT* considered to be older; as a result,
    the first commit will be done only after the first change."""
    r = git.Repository(repo)
    if r.head_is_unborn:
        return False
    now = datetime.utcnow()
    return datetime.utcfromtimestamp(
        r.head.peel().commit_time) + duration < now


def do_commit():
    """To be called in a non-daemon thread to reduce possibilities of
    early termination.

    1. Commit if
       * there is anything uncommitted, or
       * more than FORCE_AFTER_INTERVALS intervals have passed since the
         most recent commit.
       * Normal (non-forced) commits are suspended while a merge (=manual)
         operation is in progress.
    2. Timestamp using HTTPS (synchronous)
    3. (Optionally) push
    4. (Optionally) cross-timestamp using email (asynchronous), if the previous
       email has been sent more than FORCE_AFTER_INTERVALS ago. The response
       will be added to a future commit."""
    logging.start("do_commit", suffix="Threads: " +
                  str(threading.enumerate()), level=signale.XDEBUG)
    # Allow 5% of an interval tolerance, such that small timing differences
    # will not lead to lengthening the duration by one commit_interval.
    # This is as early as possible, because mail timestamps will be delayed for
    # number_of_timestampers * zeitgitter_sleep + connection_plus_work_delays
    force_interval = (autoblockchainify.config.arg.commit_interval
                      * (autoblockchainify.config.arg.force_after_intervals - 0.95))
    try:
        repo = autoblockchainify.config.arg.repository
        # If a merge (a manual process on the repository) is detected,
        # try to not interfere with the manual process and wait for the
        # next forced update
        if ((has_user_changes(repo) and not pending_merge(repo))
                or head_older_than(repo, force_interval)
                or autoblockchainify.mail.needs_timestamp()):
            # 1. Commit
            commit_current_state(repo)

            # 2. Timestamp (synchronously) using Zeitgitter
            repositories = autoblockchainify.config.arg.push_repository
            branches = autoblockchainify.config.arg.push_branch
            first = True
            for r in autoblockchainify.config.arg.zeitgitter_servers:
                if first:
                    first = False
                else:
                    time.sleep(
                        autoblockchainify.config.arg.zeitgitter_sleep.total_seconds())
                logging.pending("Timestamping with %s" %
                                r, level=signale.DEBUG)
                if '=' in r:
                    (branch, server) = r.split('=', 1)
                    cross_timestamp(repo, ['--branch', branch,
                                           '--server', server], server)
                else:
                    cross_timestamp(repo, ['--server', r], r)

            # 3. Push
            for r in repositories:
                logging.pending("Pushing upstream to %s" % r)
                push_upstream(repo, r, branches)

            # 4. Timestamp by mail (asynchronously)
            if autoblockchainify.config.arg.stamper_own_address:
                autoblockchainify.mail.async_email_timestamp()

        logging.complete("do_commit done at " +
                         datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC'))
    except Exception:
        logging.exception("Unhandled exception in commit thread")


def loop():
    """Run at given interval and offset"""
    interval = autoblockchainify.config.arg.commit_interval.total_seconds()
    offset = autoblockchainify.config.arg.commit_offset.total_seconds()
    while True:
        now = time.time()
        until = now - (now % interval) + offset
        if until <= now:
            until += interval
        time.sleep(until - now)
        threading.Thread(target=do_commit, name="commit", daemon=False).start()
